package com.my;

import java.util.*;
import java.io.*;
import javax.xml.messaging.*;
import javax.xml.soap.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;

/**
JAXM servlet that transforms and forwards shipping requests.

@author Will Provost
*/
public class ShippingAdapter 
    extends JAXMServlet 
    implements ReqRespListener
{
    /**
    Handles the SOAP message for all five operations, parsing for arguments,
    performing the operation, and returning the result.
    */
    public SOAPMessage onMessage (SOAPMessage request) 
    {
        try 
        {
            // ------------------------------------------------------------
            // Transform the request with JAXP and XSLT:
            request = transform (request, "/Incoming.xsl");
            
            // ------------------------------------------------------------
            // Echo the request, forward the call, echo the response:
            
            System.out.println ("Request to internal service:");
            request.writeTo (System.out);
            System.out.println ();

            SOAPConnectionFactory cFactory = 
                SOAPConnectionFactory.newInstance ();
            SOAPConnection connection = cFactory.createConnection ();
            SOAPMessage response = connection.call (request, serviceURI);

            System.out.println ("Response from internal service:");
            response.writeTo (System.out);
            System.out.println ();

            // ------------------------------------------------------------
            // Transform the response with JAXP and XSLT,
            //   and then with SAAJ using recursive function:
            response = transform (response, "/Outgoing.xsl");

            SOAPEnvelope element = response.getSOAPPart ().getEnvelope ();
            fix (element, element);
            response.saveChanges ();
            
            return response;
        }
        catch (Exception ex)
        {
            ex.printStackTrace ();
            return createFault (ex.getClass ().getName () + " saying \"" +
                ex.getMessage () + "\"");
        }
    }

    /**
    Transform a SOAP message based on an XSLT transform deployed with the
    application.
    */
    private SOAPMessage transform (SOAPMessage message, String transformSource)
        throws SOAPException, TransformerException, IOException,
            TransformerConfigurationException
    {
        Transformer xform = TransformerFactory.newInstance ()
            .newTransformer (new StreamSource 
                (getServletContext ().getResourceAsStream (transformSource)));

        ByteArrayOutputStream out = new ByteArrayOutputStream ();
        message.writeTo (out);
        ByteArrayInputStream in = 
            new ByteArrayInputStream (out.toByteArray ());

        MessageFactory factory = MessageFactory.newInstance ();
        MimeHeaders headers = new MimeHeaders ();
        headers.addHeader ("content-type", "text/xml");
        out = new ByteArrayOutputStream ();
        xform.transform (new StreamSource (in), new StreamResult (out));
        in = new ByteArrayInputStream (out.toByteArray ());
        return factory.createMessage (headers, in);
    }
    
    /**
    Recursive function to fix up response content.
    Transforms <b>history</b> elements to <b>trackingNumbers</b> elements.
    (More to do here to fix <b>shipment</b> elements as well.)
    */
    private SOAPElement fix (SOAPElement element, SOAPEnvelope env)
        throws SOAPException
    {
        // Recurse over child elements, but replace <history> elements 
        //   with <trackingNumbers>:
        Iterator children = element.getChildElements ();
        while (children.hasNext ())
        {
            Node node = (Node) children.next ();
            if (node instanceof SOAPElement)
            {
                SOAPElement child = (SOAPElement) node;
                if (child.getElementName ().getLocalName ().equals ("history"))
                {
                    StringTokenizer tokenizer = 
                        new StringTokenizer (child.getValue (), " :");

                    child.detachNode ();
                    
                    SOAPElement trackingNumbers = 
                        element.addChildElement ("trackingNumbers");
                    Tokens: while (tokenizer.hasMoreTokens ())
                    {
                        String shipper = tokenizer.nextToken ();
                        if (!tokenizer.hasMoreTokens ())
                            break Tokens;
                            
                        SOAPElement trackingNumber = 
                            trackingNumbers.addChildElement ("trackingNumber");
                        trackingNumber.addChildElement ("shipperID")
                            .addTextNode (shipper);
                        trackingNumber.addChildElement ("shipmentNumber")
                            .addTextNode (tokenizer.nextToken ());
                    }
                }
                else fix (child, env);
            }
        }
        
        return element;
    }
    
    /**
    Helper to create a SOAP fault message from an error string.
    */
    private SOAPMessage createFault (String message)
    {
        try
        {
            MessageFactory msgFactory = MessageFactory.newInstance ();
            SOAPMessage response = msgFactory.createMessage ();
            SOAPEnvelope envelope = response.getSOAPPart ().getEnvelope ();
            envelope.addNamespaceDeclaration ("ship", inNS);

            SOAPElement fault = envelope.getBody ()
                .addChildElement ("Fault", "soap-env");
            fault.addChildElement ("faultcode").addTextNode ("soap-env:Server");
            fault.addChildElement ("faultstring").addTextNode (message);

            return response;
        }
        catch (SOAPException ex)
        {
            System.out.println ("Exception trying to create SOAP fault.");
            return null;
        }
    }

    /**
    Hardcoded type namespace for external messages.
    */
    private static final String exNS = "http://tracking.org/messages";
        
    /**
    Hardcoded type namespace for internal messages.
    */
    private static final String inNS = "http://my.com/shipping";
    
    /**
    */
    private static final String serviceURI = 
        "http://LOCALHOST:7001/Shipping/Shipping";
}

